@@ -2,7 +2,7 @@
 
   compar.c -
 
-  $Author: nobu $
+  $Author: usa $
   created at: Thu Aug 26 14:39:48 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -31,10 +31,17 @@ rb_cmperr(VALUE x, VALUE y)
 	     rb_obj_classname(x), classname);
 }
 
+static VALUE
+cmp_eq_recursive(VALUE arg1, VALUE arg2, int recursive)
+{
+    if (recursive) return Qfalse;
+    return rb_funcall2(arg1, cmp, 1, &arg2);
+}
+
 static VALUE
 cmp_eq(VALUE *a)
 {
-    VALUE c = rb_funcall(a[0], cmp, 1, a[1]);
+    VALUE c = rb_exec_recursive_paired_outer(cmp_eq_recursive, a[0], a[1], a[1]);
 
     if (NIL_P(c)) return Qfalse;
     if (rb_cmpint(c, a[0], a[1]) == 0) return Qtrue;
@@ -49,8 +56,8 @@ cmp_failed(void)
 
 /*
  *  call-seq:
- *     obj == other    => true or false
- *  
+ *     obj == other    -> true or false
+ *
  *  Compares two objects based on the receiver's <code><=></code>
  *  method, returning true if it returns 0. Also returns true if
  *  _obj_ and _other_ are the same object.
@@ -69,8 +76,8 @@ cmp_equal(VALUE x, VALUE y)
 
 /*
  *  call-seq:
- *     obj > other    => true or false
- *  
+ *     obj > other    -> true or false
+ *
  *  Compares two objects based on the receiver's <code><=></code>
  *  method, returning true if it returns 1.
  */
@@ -86,8 +93,8 @@ cmp_gt(VALUE x, VALUE y)
 
 /*
  *  call-seq:
- *     obj >= other    => true or false
- *  
+ *     obj >= other    -> true or false
+ *
  *  Compares two objects based on the receiver's <code><=></code>
  *  method, returning true if it returns 0 or 1.
  */
@@ -103,8 +110,8 @@ cmp_ge(VALUE x, VALUE y)
 
 /*
  *  call-seq:
- *     obj < other    => true or false
- *  
+ *     obj < other    -> true or false
+ *
  *  Compares two objects based on the receiver's <code><=></code>
  *  method, returning true if it returns -1.
  */
@@ -120,8 +127,8 @@ cmp_lt(VALUE x, VALUE y)
 
 /*
  *  call-seq:
- *     obj <= other    => true or false
- *  
+ *     obj <= other    -> true or false
+ *
  *  Compares two objects based on the receiver's <code><=></code>
  *  method, returning true if it returns -1 or 0.
  */
@@ -137,17 +144,17 @@ cmp_le(VALUE x, VALUE y)
 
 /*
  *  call-seq:
- *     obj.between?(min, max)    => true or false
- *  
+ *     obj.between?(min, max)    -> true or false
+ *
  *  Returns <code>false</code> if <i>obj</i> <code><=></code>
  *  <i>min</i> is less than zero or if <i>anObject</i> <code><=></code>
  *  <i>max</i> is greater than zero, <code>true</code> otherwise.
- *     
+ *
  *     3.between?(1, 5)               #=> true
  *     6.between?(1, 5)               #=> false
  *     'cat'.between?('ant', 'dog')   #=> true
  *     'gnu'.between?('ant', 'dog')   #=> false
- *     
+ *
  */
 
 static VALUE
@@ -163,11 +170,13 @@ cmp_between(VALUE x, VALUE min, VALUE max)
  *  may be ordered. The class must define the <code><=></code> operator,
  *  which compares the receiver against another object, returning -1, 0,
  *  or +1 depending on whether the receiver is less than, equal to, or
- *  greater than the other object. <code>Comparable</code> uses
+ *  greater than the other object. If the other object is not comparable
+ *  then the <code><=></code> operator should return nil.
+ *  <code>Comparable</code> uses
  *  <code><=></code> to implement the conventional comparison operators
  *  (<code><</code>, <code><=</code>, <code>==</code>, <code>>=</code>,
  *  and <code>></code>) and the method <code>between?</code>.
- *     
+ *
  *     class SizeMatters
  *       include Comparable
  *       attr :str
@@ -181,24 +190,25 @@ cmp_between(VALUE x, VALUE min, VALUE max)
  *         @str
  *       end
  *     end
- *     
+ *
  *     s1 = SizeMatters.new("Z")
  *     s2 = SizeMatters.new("YY")
  *     s3 = SizeMatters.new("XXX")
  *     s4 = SizeMatters.new("WWWW")
  *     s5 = SizeMatters.new("VVVVV")
- *     
+ *
  *     s1 < s2                       #=> true
  *     s4.between?(s1, s3)           #=> false
  *     s4.between?(s3, s5)           #=> true
  *     [ s3, s2, s5, s4, s1 ].sort   #=> [Z, YY, XXX, WWWW, VVVVV]
- *     
+ *
  */
 
 void
 Init_Comparable(void)
 {
 #undef rb_intern
+#define rb_intern(str) rb_intern_const(str)
 
     rb_mComparable = rb_define_module("Comparable");
     rb_define_method(rb_mComparable, "==", cmp_equal, 1);
