@@ -1,5 +1,5 @@
 # $RoughId: extconf.rb,v 1.3 2001/08/14 19:54:51 knu Exp $
-# $Id: extconf.rb,v 1.7 2006/05/25 23:43:28 nobu Exp $
+# $Id: extconf.rb 25189 2009-10-02 12:04:37Z akr $
 
 require "mkmf"
 
@@ -20,10 +20,6 @@ end
 
 have_header("sys/cdefs.h")
 
-have_header("inttypes.h")
-
-have_header("unistd.h")
-
 $preload = %w[digest]
 
 create_makefile("digest/md5")
