@@ -1,4 +1,4 @@
-/* $Id: sha1ossl.h,v 1.2 2005/08/30 10:43:16 gotoyuzo Exp $ */
+/* $Id: sha1ossl.h 25189 2009-10-02 12:04:37Z akr $ */
 
 #ifndef SHA1OSSL_H_INCLUDED
 #define SHA1OSSL_H_INCLUDED
@@ -8,10 +8,13 @@
 
 #define SHA1_CTX	SHA_CTX
 
+#ifdef SHA_BLOCK_LENGTH
 #define SHA1_BLOCK_LENGTH	SHA_BLOCK_LENGTH
+#else
+#define SHA1_BLOCK_LENGTH	SHA_CBLOCK
+#endif
 #define SHA1_DIGEST_LENGTH	SHA_DIGEST_LENGTH
 
-char *SHA1_End(SHA1_CTX *ctx, char *buf);
-int SHA1_Equal(SHA1_CTX *pctx1, SHA1_CTX *pctx2);
+void SHA1_Finish(SHA1_CTX *ctx, char *buf);
 
 #endif
