@@ -1,10 +1,13 @@
+# -*- coding: euc-jp -*-
 #
 # widget demo prompts the user to select a color (called by 'widget')
 #
+#  Note: don't support ttk_wrapper. work with standard widgets only.
+#
 
 # toplevel widget ¸ߤк
 if defined?($clrpick_demo) && $clrpick_demo
-  $clrpick_demo.destroy 
+  $clrpick_demo.destroy
   $clrpick_demo = nil
 end
 
@@ -16,12 +19,15 @@ $clrpick_demo = TkToplevel.new {|w|
 }
 
 # label 
-TkLabel.new($clrpick_demo,'font'=>$font,'wraplength'=>'4i','justify'=>'left',
+#TkLabel.new($clrpick_demo,'font'=>$font,'wraplength'=>'4i','justify'=>'left',
+Tk::Label.new($clrpick_demo,'font'=>$font,'wraplength'=>'4i','justify'=>'left',
             'text'=>"ʲΥܥ򲡤ơΥɥˤ륦åȤʿطʿ򤷤Ʋ").pack('side'=>'top')
 
 # frame 
-TkFrame.new($clrpick_demo) {|frame|
-  TkButton.new(frame) {
+# TkFrame.new($clrpick_demo) {|frame|
+Tk::Frame.new($clrpick_demo) {|frame|
+  # TkButton.new(frame) {
+  Tk::Button.new(frame) {
     #text 'λ'
     text 'Ĥ'
     command proc{
@@ -31,20 +37,23 @@ TkFrame.new($clrpick_demo) {|frame|
     }
   }.pack('side'=>'left', 'expand'=>'yes')
 
-  TkButton.new(frame) {
+  # TkButton.new(frame) {
+  Tk::Button.new(frame) {
     text 'ɻ'
     command proc{showCode 'clrpick'}
   }.pack('side'=>'left', 'expand'=>'yes')
 }.pack('side'=>'bottom', 'fill'=>'x', 'pady'=>'2m')
 
 # button 
-TkButton.new($clrpick_demo, 'text'=>'طʿ ...') {|b|
-  command(proc{setColor $clrpick_demo, b, 'background', 
+# TkButton.new($clrpick_demo, 'text'=>'طʿ ...') {|b|
+Tk::Button.new($clrpick_demo, 'text'=>'طʿ ...') {|b|
+  command(proc{setColor $clrpick_demo, b, 'background',
               ['background', 'highlightbackground']})
   pack('side'=>'top', 'anchor'=>'c', 'pady'=>'2m')
 }
 
-TkButton.new($clrpick_demo, 'text'=>'ʿ ...') {|b|
+# TkButton.new($clrpick_demo, 'text'=>'ʿ ...') {|b|
+Tk::Button.new($clrpick_demo, 'text'=>'ʿ ...') {|b|
   command(proc{setColor $clrpick_demo, b, 'foreground', ['foreground']})
   pack('side'=>'top', 'anchor'=>'c', 'pady'=>'2m')
 }
@@ -52,7 +61,7 @@ TkButton.new($clrpick_demo, 'text'=>'ʿ ...') {|b|
 def setColor(w,button,name,options)
   w.grab
   initialColor = button[name]
-  color = Tk.chooseColor('title'=>"Choose a #{name} color", 'parent'=>w, 
+  color = Tk.chooseColor('title'=>"Choose a #{name} color", 'parent'=>w,
                          'initialcolor'=>initialColor)
   if color != ""
     setColor_helper(w,options,color)
