@@ -1,7 +1,7 @@
 #
 # extconf.rb
 #
-# $Id: extconf.rb,v 1.3 2006/01/09 11:00:40 ocean Exp $
+# $Id: extconf.rb 36939 2012-09-09 14:01:53Z naruse $
 #
 
 require 'mkmf'
@@ -10,7 +10,7 @@ require 'rbconfig'
 dir_config 'zlib'
 
 
-if %w'z libz zlib zdll'.find {|z| have_library(z, 'deflateReset')} and
+if %w'z libz zlib1 zlib zdll'.find {|z| have_library(z, 'deflateReset')} and
     have_header('zlib.h') then
 
   defines = []
@@ -22,7 +22,7 @@ if %w'z libz zlib zdll'.find {|z| have_library(z, 'deflateReset')} and
       os_code = 'AMIGA'
     when /\Aos2[\-_]emx\z/ then
       os_code = 'OS2'
-    when 'mswin32', 'mingw32', 'bccwin32' then
+    when /mswin|mingw|bccwin/ then
       # NOTE: cygwin should be regarded as Unix.
       os_code = 'WIN32'
     else
@@ -48,18 +48,16 @@ if %w'z libz zlib zdll'.find {|z| have_library(z, 'deflateReset')} and
     'OS_UNKNOWN' => 'Unknown',
   }
   unless OS_NAMES.key? os_code then
-    puts "invalid OS_CODE `#{os_code}'"
-    exit
+    raise "invalid OS_CODE `#{os_code}'"
   end
   message "#{OS_NAMES[os_code]}\n"
   defines << "OS_CODE=#{os_code}"
 
-  defines = defines.collect{|d|' -D'+d}.join
-  if $CPPFLAGS then
-    $CPPFLAGS += defines
-  else
-    $CFLAGS += defines
-  end
+  $defs.concat(defines.collect{|d|' -D'+d})
+
+  have_func('crc32_combine', 'zlib.h')
+  have_func('adler32_combine', 'zlib.h')
+  have_type('z_crc_t', 'zlib.h')
 
   create_makefile('zlib')
 
