@@ -4,15 +4,15 @@
 #++
 # Copyright (c) 2002-2008 Akinori MUSHA <knu@iDaemons.org>
 #
-# Documentation by Akinori MUSHA and Gavin Sinclair. 
+# Documentation by Akinori MUSHA and Gavin Sinclair.
 #
 # All rights reserved.  You can redistribute and/or modify it under the same
 # terms as Ruby.
 #
-#   $Id: set.rb 17045 2008-06-09 08:58:30Z knu $
+#   $Id: set.rb 34533 2012-02-10 08:28:48Z naruse $
+#
+# == Overview
 #
-# == Overview 
-# 
 # This library provides the Set class, which deals with a collection
 # of unordered values with no duplicates.  It is a hybrid of Array's
 # intuitive inter-operation facilities and Hash's fast lookup.  If you
@@ -20,7 +20,7 @@
 #
 # The method +to_set+ is added to Enumerable for convenience.
 #
-# See the Set class for an example of usage.
+# See the Set and SortedSet documentation for examples of usage.
 
 
 #
@@ -70,30 +70,41 @@ class Set
     enum.nil? and return
 
     if block
-      enum.each { |o| add(block[o]) }
+      do_with_enum(enum) { |o| add(block[o]) }
     else
       merge(enum)
     end
   end
 
+  def do_with_enum(enum, &block) # :nodoc:
+    if enum.respond_to?(:each_entry)
+      enum.each_entry(&block)
+    elsif enum.respond_to?(:each)
+      enum.each(&block)
+    else
+      raise ArgumentError, "value must be enumerable"
+    end
+  end
+  private :do_with_enum
+
   # Copy internal hash.
   def initialize_copy(orig)
     @hash = orig.instance_eval{@hash}.dup
   end
 
-  def freeze	# :nodoc:
+  def freeze    # :nodoc:
     super
     @hash.freeze
     self
   end
 
-  def taint	# :nodoc:
+  def taint     # :nodoc:
     super
     @hash.taint
     self
   end
 
-  def untaint	# :nodoc:
+  def untaint   # :nodoc:
     super
     @hash.untaint
     self
@@ -119,11 +130,11 @@ class Set
   # Replaces the contents of the set with the contents of the given
   # enumerable object and returns self.
   def replace(enum)
-    if enum.class == self.class
-      @hash.replace(enum.instance_eval { @hash })
+    if enum.instance_of?(self.class)
+      @hash.replace(enum.instance_variable_get(:@hash))
     else
       clear
-      enum.each { |o| add(o) }
+      merge(enum)
     end
 
     self
@@ -134,18 +145,18 @@ class Set
     @hash.keys
   end
 
-  def flatten_merge(set, seen = Set.new)
+  def flatten_merge(set, seen = Set.new) # :nodoc:
     set.each { |e|
       if e.is_a?(Set)
-	if seen.include?(e_id = e.object_id)
-	  raise ArgumentError, "tried to flatten recursive Set"
-	end
+        if seen.include?(e_id = e.object_id)
+          raise ArgumentError, "tried to flatten recursive Set"
+        end
 
-	seen.add(e_id)
-	flatten_merge(e, seen)
-	seen.delete(e_id)
+        seen.add(e_id)
+        flatten_merge(e, seen)
+        seen.delete(e_id)
       else
-	add(e)
+        add(e)
       end
     }
 
@@ -255,6 +266,14 @@ class Set
     self
   end
 
+  # Deletes every element of the set for which block evaluates to
+  # false, and returns self.
+  def keep_if
+    block_given? or return enum_for(__method__)
+    to_a.each { |o| @hash.delete(o) unless yield(o) }
+    self
+  end
+
   # Replaces the elements with ones returned by collect().
   def collect!
     block_given? or return enum_for(__method__)
@@ -273,13 +292,22 @@ class Set
     size == n ? nil : self
   end
 
+  # Equivalent to Set#keep_if, but returns nil if no changes were
+  # made.
+  def select!
+    block_given? or return enum_for(__method__)
+    n = size
+    keep_if { |o| yield(o) }
+    size == n ? nil : self
+  end
+
   # Merges the elements of the given enumerable object to the set and
   # returns self.
   def merge(enum)
-    if enum.is_a?(Set)
-      @hash.update(enum.instance_eval { @hash })
+    if enum.instance_of?(self.class)
+      @hash.update(enum.instance_variable_get(:@hash))
     else
-      enum.each { |o| add(o) }
+      do_with_enum(enum) { |o| add(o) }
     end
 
     self
@@ -288,7 +316,7 @@ class Set
   # Deletes every element that appears in the given enumerable object
   # and returns self.
   def subtract(enum)
-    enum.each { |o| delete(o) }
+    do_with_enum(enum) { |o| delete(o) }
     self
   end
 
@@ -297,24 +325,24 @@ class Set
   def |(enum)
     dup.merge(enum)
   end
-  alias + |		##
-  alias union |		##
+  alias + |             ##
+  alias union |         ##
 
   # Returns a new set built by duplicating the set, removing every
   # element that appears in the given enumerable object.
   def -(enum)
     dup.subtract(enum)
   end
-  alias difference -	##
+  alias difference -    ##
 
   # Returns a new set containing elements common to the set and the
   # given enumerable object.
   def &(enum)
     n = self.class.new
-    enum.each { |o| n.add(o) if include?(o) }
+    do_with_enum(enum) { |o| n.add(o) if include?(o) }
     n
   end
-  alias intersection &	##
+  alias intersection &  ##
 
   # Returns a new set containing elements exclusive between the set
   # and the given enumerable object.  (set ^ enum) is equivalent to
@@ -327,20 +355,23 @@ class Set
 
   # Returns true if two sets are equal.  The equality of each couple
   # of elements is defined according to Object#eql?.
-  def ==(set)
-    equal?(set) and return true
-
-    set.is_a?(Set) && size == set.size or return false
-
-    hash = @hash.dup
-    set.all? { |o| hash.include?(o) }
+  def ==(other)
+    if self.equal?(other)
+      true
+    elsif other.instance_of?(self.class)
+      @hash == other.instance_variable_get(:@hash)
+    elsif other.is_a?(Set) && self.size == other.size
+      other.all? { |o| @hash.include?(o) }
+    else
+      false
+    end
   end
 
-  def hash	# :nodoc:
+  def hash      # :nodoc:
     @hash.hash
   end
 
-  def eql?(o)	# :nodoc:
+  def eql?(o)   # :nodoc:
     return false unless o.is_a?(Set)
     @hash.eql?(o.instance_eval{@hash})
   end
@@ -393,23 +424,23 @@ class Set
     if func.arity == 2
       require 'tsort'
 
-      class << dig = {}		# :nodoc:
-	include TSort
+      class << dig = {}         # :nodoc:
+        include TSort
 
-	alias tsort_each_node each_key
-	def tsort_each_child(node, &block)
-	  fetch(node).each(&block)
-	end
+        alias tsort_each_node each_key
+        def tsort_each_child(node, &block)
+          fetch(node).each(&block)
+        end
       end
 
       each { |u|
-	dig[u] = a = []
-	each{ |v| func.call(u, v) and a << v }
+        dig[u] = a = []
+        each{ |v| func.call(u, v) and a << v }
       }
 
       set = Set.new()
       dig.each_strongly_connected_component { |css|
-	set.add(self.class.new(css))
+        set.add(self.class.new(css))
       }
       set
     else
@@ -436,108 +467,153 @@ class Set
     end
   end
 
-  def pretty_print(pp)	# :nodoc:
+  def pretty_print(pp)  # :nodoc:
     pp.text sprintf('#<%s: {', self.class.name)
     pp.nest(1) {
       pp.seplist(self) { |o|
-	pp.pp o
+        pp.pp o
       }
     }
     pp.text "}>"
   end
 
-  def pretty_print_cycle(pp)	# :nodoc:
+  def pretty_print_cycle(pp)    # :nodoc:
     pp.text sprintf('#<%s: {%s}>', self.class.name, empty? ? '' : '...')
   end
 end
 
-# SortedSet implements a set which elements are sorted in order.  See Set.
+#
+# SortedSet implements a Set that guarantees that it's element are
+# yielded in sorted order (according to the return values of their
+# #<=> methods) when iterating over them.
+#
+# All elements that are added to a SortedSet must respond to the <=>
+# method for comparison.
+#
+# Also, all elements must be <em>mutually comparable</em>: <tt>el1 <=>
+# el2</tt> must not return <tt>nil</tt> for any elements <tt>el1</tt>
+# and <tt>el2</tt>, else an ArgumentError will be raised when
+# iterating over the SortedSet.
+#
+# == Example
+#
+#   require "set"
+#
+#   set = SortedSet.new([2, 1, 5, 6, 4, 5, 3, 3, 3])
+#   ary = []
+#
+#   set.each do |obj|
+#     ary << obj
+#   end
+#
+#   p ary # => [1, 2, 3, 4, 5, 6]
+#
+#   set2 = SortedSet.new([1, 2, "3"])
+#   set2.each { |obj| } # => raises ArgumentError: comparison of Fixnum with String failed
+#
 class SortedSet < Set
   @@setup = false
 
   class << self
-    def [](*ary)	# :nodoc:
+    def [](*ary)        # :nodoc:
       new(ary)
     end
 
-    def setup	# :nodoc:
+    def setup   # :nodoc:
       @@setup and return
 
       module_eval {
         # a hack to shut up warning
         alias old_init initialize
-        remove_method :old_init
       }
       begin
-	require 'rbtree'
+        require 'rbtree'
 
-	module_eval %{
-	  def initialize(*args, &block)
-	    @hash = RBTree.new
-	    super
-	  end
-	}
+        module_eval %{
+          def initialize(*args, &block)
+            @hash = RBTree.new
+            super
+          end
+
+          def add(o)
+            o.respond_to?(:<=>) or raise ArgumentError, "value must respond to <=>"
+            super
+          end
+          alias << add
+        }
       rescue LoadError
-	module_eval %{
-	  def initialize(*args, &block)
-	    @keys = nil
-	    super
-	  end
+        module_eval %{
+          def initialize(*args, &block)
+            @keys = nil
+            super
+          end
 
-	  def clear
-	    @keys = nil
-	    super
-	  end
+          def clear
+            @keys = nil
+            super
+          end
 
-	  def replace(enum)
-	    @keys = nil
-	    super
-	  end
+          def replace(enum)
+            @keys = nil
+            super
+          end
 
-	  def add(o)
-	    @keys = nil
-	    @hash[o] = true
-	    self
-	  end
-	  alias << add
+          def add(o)
+            o.respond_to?(:<=>) or raise ArgumentError, "value must respond to <=>"
+            @keys = nil
+            super
+          end
+          alias << add
 
-	  def delete(o)
-	    @keys = nil
-	    @hash.delete(o)
-	    self
-	  end
+          def delete(o)
+            @keys = nil
+            @hash.delete(o)
+            self
+          end
 
-	  def delete_if
+          def delete_if
             block_given? or return enum_for(__method__)
-	    n = @hash.size
-	    super
-	    @keys = nil if @hash.size != n
-	    self
-	  end
+            n = @hash.size
+            super
+            @keys = nil if @hash.size != n
+            self
+          end
 
-	  def merge(enum)
-	    @keys = nil
-	    super
-	  end
+          def keep_if
+            block_given? or return enum_for(__method__)
+            n = @hash.size
+            super
+            @keys = nil if @hash.size != n
+            self
+          end
 
-	  def each
-	    block_given? or return enum_for(__method__)
-	    to_a.each { |o| yield(o) }
-	    self
-	  end
+          def merge(enum)
+            @keys = nil
+            super
+          end
 
-	  def to_a
-	    (@keys = @hash.keys).sort! unless @keys
-	    @keys
-	  end
-	}
+          def each
+            block_given? or return enum_for(__method__)
+            to_a.each { |o| yield(o) }
+            self
+          end
+
+          def to_a
+            (@keys = @hash.keys).sort! unless @keys
+            @keys
+          end
+        }
       end
+      module_eval {
+        # a hack to shut up warning
+        remove_method :old_init
+      }
 
       @@setup = true
     end
   end
 
-  def initialize(*args, &block)	# :nodoc:
+  def initialize(*args, &block) # :nodoc:
     SortedSet.setup
     initialize(*args, &block)
   end
@@ -555,87 +631,89 @@ end
 # == RestricedSet class
 # RestricedSet implements a set with restrictions defined by a given
 # block.
-# 
+#
 # === Super class
 #     Set
-# 
+#
 # === Class Methods
 # --- RestricedSet::new(enum = nil) { |o| ... }
 # --- RestricedSet::new(enum = nil) { |rset, o| ... }
 #     Creates a new restricted set containing the elements of the given
 #     enumerable object.  Restrictions are defined by the given block.
-# 
+#
 #     If the block's arity is 2, it is called with the RestrictedSet
 #     itself and an object to see if the object is allowed to be put in
 #     the set.
-# 
+#
 #     Otherwise, the block is called with an object to see if the object
 #     is allowed to be put in the set.
-# 
+#
 # === Instance Methods
 # --- restriction_proc
 #     Returns the restriction procedure of the set.
-# 
+#
 # =end
-# 
+#
 # class RestricedSet < Set
 #   def initialize(*args, &block)
 #     @proc = block or raise ArgumentError, "missing a block"
-# 
+#
 #     if @proc.arity == 2
 #       instance_eval %{
-# 	def add(o)
-# 	  @hash[o] = true if @proc.call(self, o)
-# 	  self
-# 	end
-# 	alias << add
-# 
-# 	def add?(o)
-# 	  if include?(o) || !@proc.call(self, o)
-# 	    nil
-# 	  else
-# 	    @hash[o] = true
-# 	    self
-# 	  end
-# 	end
-# 
-# 	def replace(enum)
-# 	  clear
-# 	  enum.each { |o| add(o) }
-# 
-# 	  self
-# 	end
-# 
-# 	def merge(enum)
-# 	  enum.each { |o| add(o) }
-# 
-# 	  self
-# 	end
+#       def add(o)
+#         @hash[o] = true if @proc.call(self, o)
+#         self
+#       end
+#       alias << add
+#
+#       def add?(o)
+#         if include?(o) || !@proc.call(self, o)
+#           nil
+#         else
+#           @hash[o] = true
+#           self
+#         end
+#       end
+#
+#       def replace(enum)
+#         enum.respond_to?(:each) or raise ArgumentError, "value must be enumerable"
+#         clear
+#         enum.each_entry { |o| add(o) }
+#
+#         self
+#       end
+#
+#       def merge(enum)
+#         enum.respond_to?(:each) or raise ArgumentError, "value must be enumerable"
+#         enum.each_entry { |o| add(o) }
+#
+#         self
+#       end
 #       }
 #     else
 #       instance_eval %{
-# 	def add(o)
+#       def add(o)
 #         if @proc.call(o)
-# 	    @hash[o] = true 
+#           @hash[o] = true
 #         end
-# 	  self
-# 	end
-# 	alias << add
-# 
-# 	def add?(o)
-# 	  if include?(o) || !@proc.call(o)
-# 	    nil
-# 	  else
-# 	    @hash[o] = true
-# 	    self
-# 	  end
-# 	end
+#         self
+#       end
+#       alias << add
+#
+#       def add?(o)
+#         if include?(o) || !@proc.call(o)
+#           nil
+#         else
+#           @hash[o] = true
+#           self
+#         end
+#       end
 #       }
 #     end
-# 
+#
 #     super(*args)
 #   end
-# 
+#
 #   def restriction_proc
 #     @proc
 #   end
@@ -674,10 +752,10 @@ class TC_Set < Test::Unit::TestCase
       Set.new([1,2])
       Set.new('a'..'c')
     }
-    assert_raises(NoMethodError) {
+    assert_raises(ArgumentError) {
       Set.new(false)
     }
-    assert_raises(NoMethodError) {
+    assert_raises(ArgumentError) {
       Set.new(1)
     }
     assert_raises(ArgumentError) {
@@ -760,12 +838,12 @@ class TC_Set < Test::Unit::TestCase
     set1 = Set[
       1,
       Set[
-	5,
-	Set[7,
-	  Set[0]
-	],
-	Set[6,2],
-	1
+        5,
+        Set[7,
+          Set[0]
+        ],
+        Set[6,2],
+        1
       ],
       3,
       Set[3,4]
@@ -941,11 +1019,11 @@ class TC_Set < Test::Unit::TestCase
     assert_same(set, ret)
 
     e = set.each
-    assert_instance_of(Enumerable::Enumerator, e)
+    assert_instance_of(Enumerator, e)
 
     assert_nothing_raised {
       set.each { |o|
-	ary.delete(o) or raise "unexpected element: #{o}"
+        ary.delete(o) or raise "unexpected element: #{o}"
       }
 
       ary.empty? or raise "forgotten elements: #{ary.join(', ')}"
@@ -1010,11 +1088,11 @@ class TC_Set < Test::Unit::TestCase
     ret = set.collect! { |i|
       case i
       when Numeric
-	i * 2
+        i * 2
       when String
-	i.upcase
+        i.upcase
       else
-	nil
+        nil
       end
     }
 
@@ -1139,15 +1217,15 @@ class TC_Set < Test::Unit::TestCase
     assert_equal(set, ret.flatten)
     ret.each { |s|
       if s.include?(0)
-	assert_equal(Set[0,1], s)
+        assert_equal(Set[0,1], s)
       elsif s.include?(3)
-	assert_equal(Set[3,4,5], s)
+        assert_equal(Set[3,4,5], s)
       elsif s.include?(7)
-	assert_equal(Set[7], s)
+        assert_equal(Set[7], s)
       elsif s.include?(9)
-	assert_equal(Set[9,10,11], s)
+        assert_equal(Set[9,10,11], s)
       else
-	raise "unexpected group: #{s.inspect}"
+        raise "unexpected group: #{s.inspect}"
       end
     }
   end
@@ -1241,33 +1319,33 @@ end
 # class TC_RestricedSet < Test::Unit::TestCase
 #   def test_s_new
 #     assert_raises(ArgumentError) { RestricedSet.new }
-# 
+#
 #     s = RestricedSet.new([-1,2,3]) { |o| o > 0 }
 #     assert_equal([2,3], s.sort)
 #   end
-# 
+#
 #   def test_restriction_proc
 #     s = RestricedSet.new([-1,2,3]) { |o| o > 0 }
-# 
+#
 #     f = s.restriction_proc
 #     assert_instance_of(Proc, f)
 #     assert(f[1])
 #     assert(!f[0])
 #   end
-# 
+#
 #   def test_replace
 #     s = RestricedSet.new(-3..3) { |o| o > 0 }
 #     assert_equal([1,2,3], s.sort)
-# 
+#
 #     s.replace([-2,0,3,4,5])
 #     assert_equal([3,4,5], s.sort)
 #   end
-# 
+#
 #   def test_merge
 #     s = RestricedSet.new { |o| o > 0 }
 #     s.merge(-5..5)
 #     assert_equal([1,2,3,4,5], s.sort)
-# 
+#
 #     s.merge([10,-10,-8,8])
 #     assert_equal([1,2,3,4,5,8,10], s.sort)
 #   end
