@@ -5,6 +5,8 @@ class TestOptionParser::OptArg < TestOptionParser
     super
     @opt.def_option("-x[VAL]") {|x| @flag = x}
     @opt.def_option("--option[=VAL]") {|x| @flag = x}
+    @opt.def_option("--regexp[=REGEXP]", Regexp) {|x| @reopt = x}
+    @reopt = nil
   end
 
   def test_short
