@@ -1,30 +1,31 @@
-require_relative 'gemutilities'
+require 'rubygems/test_case'
 require 'rubygems/doc_manager'
 
-class TestGemDocManager < RubyGemTestCase
+class TestGemDocManager < Gem::TestCase
 
   def setup
     super
 
-    @spec = quick_gem 'a'
+    @spec = quick_gem 'a', 2
     @manager = Gem::DocManager.new(@spec)
   end
 
   def test_uninstall_doc_unwritable
-    orig_mode = File.stat(@spec.installation_path).mode
+    path = @spec.base_dir
+    orig_mode = File.stat(path).mode
 
     # File.chmod has no effect on MS Windows directories (it needs ACL).
     if win_platform?
       skip("test_uninstall_doc_unwritable skipped on MS Windows")
     else
-      File.chmod(0, @spec.installation_path)
+      FileUtils.chmod 0000, path
     end
 
     assert_raises Gem::FilePermissionError do
       @manager.uninstall_doc
     end
   ensure
-    File.chmod orig_mode, @spec.installation_path
+    FileUtils.chmod orig_mode, path
   end
 
 end
