@@ -1,7 +1,7 @@
 require "test/unit"
+require_relative "utils.rb"
 require "webrick"
 require "stringio"
-require File.join(File.dirname(__FILE__), "utils.rb")
 
 class WEBrick::TestFileHandler < Test::Unit::TestCase
   def default_file_handler(filename)
@@ -22,11 +22,11 @@ class WEBrick::TestFileHandler < Test::Unit::TestCase
   end
 
   def make_range_request(range_spec)
-    msg = <<-_end_of_request_
+    msg = <<-END_OF_REQUEST
       GET / HTTP/1.0
       Range: #{range_spec}
 
-    _end_of_request_
+    END_OF_REQUEST
     return StringIO.new(msg.gsub(/^ {6}/, ""))
   end
 
@@ -75,43 +75,109 @@ class WEBrick::TestFileHandler < Test::Unit::TestCase
   def test_filehandler
     config = { :DocumentRoot => File.dirname(__FILE__), }
     this_file = File.basename(__FILE__)
-    TestWEBrick.start_httpserver(config) do |server, addr, port|
+    filesize = File.size(__FILE__)
+    this_data = File.open(__FILE__, "rb") {|f| f.read}
+    range = nil
+    bug2593 = '[ruby-dev:40030]'
+
+    TestWEBrick.start_httpserver(config) do |server, addr, port, log|
       http = Net::HTTP.new(addr, port)
       req = Net::HTTP::Get.new("/")
       http.request(req){|res|
-        assert_equal("200", res.code)
-        assert_equal("text/html", res.content_type)
-        assert_match(/HREF="#{this_file}"/, res.body)
+        assert_equal("200", res.code, log.call)
+        assert_equal("text/html", res.content_type, log.call)
+        assert_match(/HREF="#{this_file}"/, res.body, log.call)
       }
       req = Net::HTTP::Get.new("/#{this_file}")
       http.request(req){|res|
-        assert_equal("200", res.code)
-        assert_equal("text/plain", res.content_type)
-        assert_equal(File.read(__FILE__), res.body)
+        assert_equal("200", res.code, log.call)
+        assert_equal("text/plain", res.content_type, log.call)
+        assert_equal(File.read(__FILE__), res.body, log.call)
       }
+
+      req = Net::HTTP::Get.new("/#{this_file}", "range"=>"bytes=#{filesize-100}-")
+      http.request(req){|res|
+        assert_equal("206", res.code, log.call)
+        assert_equal("text/plain", res.content_type, log.call)
+        assert_nothing_raised(bug2593) {range = res.content_range}
+        assert_equal((filesize-100)..(filesize-1), range, log.call)
+        assert_equal(this_data[-100..-1], res.body, log.call)
+      }
+
+      req = Net::HTTP::Get.new("/#{this_file}", "range"=>"bytes=-100")
+      http.request(req){|res|
+        assert_equal("206", res.code, log.call)
+        assert_equal("text/plain", res.content_type, log.call)
+        assert_nothing_raised(bug2593) {range = res.content_range}
+        assert_equal((filesize-100)..(filesize-1), range, log.call)
+        assert_equal(this_data[-100..-1], res.body, log.call)
+      }
+
+      req = Net::HTTP::Get.new("/#{this_file}", "range"=>"bytes=0-99")
+      http.request(req){|res|
+        assert_equal("206", res.code, log.call)
+        assert_equal("text/plain", res.content_type, log.call)
+        assert_nothing_raised(bug2593) {range = res.content_range}
+        assert_equal(0..99, range, log.call)
+        assert_equal(this_data[0..99], res.body, log.call)
+      }
+
+      req = Net::HTTP::Get.new("/#{this_file}", "range"=>"bytes=100-199")
+      http.request(req){|res|
+        assert_equal("206", res.code, log.call)
+        assert_equal("text/plain", res.content_type, log.call)
+        assert_nothing_raised(bug2593) {range = res.content_range}
+        assert_equal(100..199, range, log.call)
+        assert_equal(this_data[100..199], res.body, log.call)
+      }
+
+      req = Net::HTTP::Get.new("/#{this_file}", "range"=>"bytes=0-0")
+      http.request(req){|res|
+        assert_equal("206", res.code, log.call)
+        assert_equal("text/plain", res.content_type, log.call)
+        assert_nothing_raised(bug2593) {range = res.content_range}
+        assert_equal(0..0, range, log.call)
+        assert_equal(this_data[0..0], res.body, log.call)
+      }
+
+      req = Net::HTTP::Get.new("/#{this_file}", "range"=>"bytes=-1")
+      http.request(req){|res|
+        assert_equal("206", res.code, log.call)
+        assert_equal("text/plain", res.content_type, log.call)
+        assert_nothing_raised(bug2593) {range = res.content_range}
+        assert_equal((filesize-1)..(filesize-1), range, log.call)
+        assert_equal(this_data[-1, 1], res.body, log.call)
+      }
+
+      req = Net::HTTP::Get.new("/#{this_file}", "range"=>"bytes=0-0, -2")
+      http.request(req){|res|
+        assert_equal("206", res.code, log.call)
+        assert_equal("multipart/byteranges", res.content_type, log.call)
+      }
+
     end
   end
 
   def test_non_disclosure_name
     config = { :DocumentRoot => File.dirname(__FILE__), }
     this_file = File.basename(__FILE__)
-    TestWEBrick.start_httpserver(config) do |server, addr, port|
+    TestWEBrick.start_httpserver(config) do |server, addr, port, log|
       http = Net::HTTP.new(addr, port)
       doc_root_opts = server[:DocumentRootOptions]
       doc_root_opts[:NondisclosureName] = %w(.ht* *~ test_*)
       req = Net::HTTP::Get.new("/")
       http.request(req){|res|
-        assert_equal("200", res.code)
-        assert_equal("text/html", res.content_type)
+        assert_equal("200", res.code, log.call)
+        assert_equal("text/html", res.content_type, log.call)
         assert_no_match(/HREF="#{File.basename(__FILE__)}"/, res.body)
       }
       req = Net::HTTP::Get.new("/#{this_file}")
       http.request(req){|res|
-        assert_equal("404", res.code)
+        assert_equal("404", res.code, log.call)
       }
       doc_root_opts[:NondisclosureName] = %w(.ht* *~ TEST_*)
       http.request(req){|res|
-        assert_equal("404", res.code)
+        assert_equal("404", res.code, log.call)
       }
     end
   end
@@ -119,14 +185,14 @@ class WEBrick::TestFileHandler < Test::Unit::TestCase
   def test_directory_traversal
     config = { :DocumentRoot => File.dirname(__FILE__), }
     this_file = File.basename(__FILE__)
-    TestWEBrick.start_httpserver(config) do |server, addr, port|
+    TestWEBrick.start_httpserver(config) do |server, addr, port, log|
       http = Net::HTTP.new(addr, port)
       req = Net::HTTP::Get.new("/../../")
-      http.request(req){|res| assert_equal("400", res.code) }
+      http.request(req){|res| assert_equal("400", res.code, log.call) }
       req = Net::HTTP::Get.new("/..%5c../#{File.basename(__FILE__)}")
-      http.request(req){|res| assert_equal(windows? ? "200" : "404", res.code) }
+      http.request(req){|res| assert_equal(windows? ? "200" : "404", res.code, log.call) }
       req = Net::HTTP::Get.new("/..%5c..%5cruby.c")
-      http.request(req){|res| assert_equal("404", res.code) }
+      http.request(req){|res| assert_equal("404", res.code, log.call) }
     end
   end
 
@@ -134,10 +200,10 @@ class WEBrick::TestFileHandler < Test::Unit::TestCase
     if windows?
       config = { :DocumentRoot => File.dirname(__FILE__), }
       this_file = File.basename(__FILE__)
-      TestWEBrick.start_httpserver(config) do |server, addr, port|
+      TestWEBrick.start_httpserver(config) do |server, addr, port, log|
         http = Net::HTTP.new(addr, port)
         req = Net::HTTP::Get.new("/..%5c..")
-        http.request(req){|res| assert_equal("301", res.code) }
+        http.request(req){|res| assert_equal("301", res.code, log.call) }
       end
     end
   end
@@ -148,25 +214,32 @@ class WEBrick::TestFileHandler < Test::Unit::TestCase
       :DocumentRoot => File.dirname(__FILE__),
       :CGIPathEnv => ENV['PATH'],
     }
-    TestWEBrick.start_httpserver(config) do |server, addr, port|
+    TestWEBrick.start_httpserver(config) do |server, addr, port, log|
       http = Net::HTTP.new(addr, port)
-
-      req = Net::HTTP::Get.new("/webric~1.cgi/test")
+      if windows?
+        fname = nil
+        Dir.chdir(config[:DocumentRoot]) do
+          fname = IO.popen("dir /x webrick_long_filename.cgi", "r").read.match(/\s(w.+?cgi)\s/i)[1].downcase
+        end
+      else
+        fname = "webric~1.cgi"
+      end
+      req = Net::HTTP::Get.new("/#{fname}/test")
       http.request(req) do |res|
         if windows?
-          assert_equal("200", res.code)
-          assert_equal("/test", res.body)
+          assert_equal("200", res.code, log.call)
+          assert_equal("/test", res.body, log.call)
         else
-          assert_equal("404", res.code)
+          assert_equal("404", res.code, log.call)
         end
       end
 
       req = Net::HTTP::Get.new("/.htaccess")
-      http.request(req) {|res| assert_equal("404", res.code) }
+      http.request(req) {|res| assert_equal("404", res.code, log.call) }
       req = Net::HTTP::Get.new("/htacce~1")
-      http.request(req) {|res| assert_equal("404", res.code) }
+      http.request(req) {|res| assert_equal("404", res.code, log.call) }
       req = Net::HTTP::Get.new("/HTACCE~1")
-      http.request(req) {|res| assert_equal("404", res.code) }
+      http.request(req) {|res| assert_equal("404", res.code, log.call) }
     end
   end
 
@@ -175,22 +248,29 @@ class WEBrick::TestFileHandler < Test::Unit::TestCase
       :CGIInterpreter => TestWEBrick::RubyBin,
       :DocumentRoot => File.dirname(__FILE__),
       :CGIPathEnv => ENV['PATH'],
+      :RequestCallback => Proc.new{|req, res|
+        def req.meta_vars
+          meta = super
+          meta["RUBYLIB"] = $:.join(File::PATH_SEPARATOR)
+          return meta
+        end
+      },
     }
-    TestWEBrick.start_httpserver(config) do |server, addr, port|
+    TestWEBrick.start_httpserver(config) do |server, addr, port, log|
       http = Net::HTTP.new(addr, port)
 
       req = Net::HTTP::Get.new("/webrick.cgi/test")
       http.request(req) do |res|
-        assert_equal("200", res.code)
-        assert_equal("/test", res.body)
+        assert_equal("200", res.code, log.call)
+        assert_equal("/test", res.body, log.call)
       end
 
       response_assertion = Proc.new do |res|
         if windows?
-          assert_equal("200", res.code)
-          assert_equal("/test", res.body)
+          assert_equal("200", res.code, log.call)
+          assert_equal("/test", res.body, log.call)
         else
-          assert_equal("404", res.code)
+          assert_equal("404", res.code, log.call)
         end
       end
       req = Net::HTTP::Get.new("/webrick.cgi%20/test")
